#
# DecoMecha - diving decompression library.
#
# Copyright (C) 2013 by Artur Wroblewski <wrobell@pld-linux.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import ctypes as ct
from .cglue import lib, libc, Profile, Segment, Conf, GasMix, raise_error
from .util import ptr_property, Array


class DecoEngine(object):
    """
    Dive decompression calculation engine.

    Example::

        >>> deco = DecoEngine()
        >>> deco.gf_low = 20
        >>> deco.add_gas(0, 27)
        >>> deco.add_gas(22, 50)
        >>> deco.add_segment(45, 30)
        >>> info = deco.calculate()
        >>> for s in info.stops: print(' {:3}m {}min'.format(s.depth, s.time))
          21m 1min
          18m 1min
          15m 1min
          12m 2min
           9m 4min
           6m 6min
           3m 13min
    """
    def __init__(self):
        """
        Create decompression engine instance.
        """
        self._conf = lib.deco_conf_new()
        self.profile = Array(ct.pointer(Profile()), 'segments', 'n_segments')
        self.gas_list = Array(self._conf, 'gas_list', 'n_gas_mixes')
        self.setpoints = Array(self._conf, 'setpoints', 'n_setpoints')


    def add_segment(self, depth, time):
        """
        Add dive profile segment.

        Multi-level dives are not supported at the moment (only one dive
        segment can be specified).

        :Parameters:
         depth
            Depth of a dive segment.
         time
            Dive time at given depth.
        """
        r = lib.deco_segment_add(self.profile._array, depth, time)
        if r == -1:
            raise_error()


    def add_gas(self, depth, o2, he=0):
        """
        Add gas mix to gas list.

        First gas depth should be 0m.

        :Parameters:
         depth
            Depth at which gas should be switched on.
         o2
            O2 perecentage, i.e. 32.
         he
            Helium percentage.
        """
        r = lib.deco_gas_add(self._conf, depth, o2, he)
        if r == -1:
            raise_error()


    def add_setpoint(self, depth, ppO2):
        """
        Add setpoint to enable CCR dive.

        First setpoint should be at depth 0m.

        Only one setpoint is supported at the moment.

        :Parameters:
         depth
            Depth at which setpoint is switched on.
         ppO2
            O2 partial pressure of the setpoint, i.e. 1.3.
        """
        r = lib.deco_setpoint_add(self._conf, depth, ppO2)
        if r == -1:
            raise_error()


    def calculate(self):
        """
        Calculate decompression information.

        Instance of :py:class:`decomecha.DecoInfo` is returned.

        .. seealso::
            :py:class:`decomecha.DecoInfo`
        """
        data = lib.deco_calculate(self.profile._array, self._conf)
        deco_info = DecoInfo()
        deco_info.stops = Array(data, 'stops', 'n_stops')
        deco_info.ndl = data.contents.ndl
        return deco_info


    def _set_altitude(self, altitude):
        """
        Set altitude for decompression calculations.
        """
        lib.deco_set_altitude(self._conf, altitude)


    def __del__(self):
        lib.deco_free(None, self._conf, None)
        libc.free(self.profile._array.contents.segments)


    altitude = property(fset=_set_altitude, doc='Altitude of a dive' \
            ' (read-only), see also `pressure`.')
    last_stop = ptr_property('_conf', 'last_stop',
            doc='Last decompression stop.')
    gf_low = ptr_property('_conf', 'gf_low',
            doc='Gradient Factor Low parameter, i.e. 30%.')
    gf_high = ptr_property('_conf', 'gf_high',
            doc='Gradient Factor High parameter, i.e. 85%.')
    pressure = ptr_property('_conf', 'pressure', doc='Pressure of a dive' \
            ' (read-write) in millibar, i.e. 1013, see also `altitude`.')



class DecoInfo(object):
    """
    Decompression information.

    :Attributes:
     stops
        List of decompression stops.
     ndl
        NDL time of non-decompression dive.
    """
    def __init__(self):
        """
        Create decompression information object.
        """
        self.stops = []
        self.ndl = 0


    def __del__(self):
        if isinstance(self.stops, Array):
            libc.free(self.stops._array)


# vim: sw=4:et:ai
