#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include "deco.h"

#include "p2_definitions.h"
#include "shared_definitions.h"

#define DECO_SIM_SPEED 30
#define DECO_METER_TO_MBAR 99.85 // from p2_deco.c
#define DECO_STD_ATM 1013.25

// error messages, see deco_error function
static char *deco_errors[] = {
    "no error",
    "unable to allocate memory",
    "feature not supported",
    "depth below maximum supported depth",
    "maximum number of gas mixes defined",
    "maximum number of setpoints defined"
};

// OSTC custom functions
unsigned short custom_functions[64];

// OSTC firmware information about tissues saturation
extern float pres_tissue_N2[NUM_COMP];
extern float pres_tissue_He[NUM_COMP];

/**
 * Create decompression info structure and copy decompression stops
 * information.
 *
 * The deco stops information is gathered from internal OSTC firmware data
 * structtures.
 */
deco_info_t *_copy_deco_stops() {
    int i, k;
    deco_info_t *di = calloc(1, sizeof(deco_info_t));

    di->stops = calloc(NUM_STOPS, sizeof(deco_stop_t));

    i = 0;
    k = 0;
    // copy OSTC internal deco stop data into deco schedule info
    while (i < NUM_STOPS && char_O_deco_depth[i] != 0) {
        // ostc keeps deco time as unsigned char, aggregate by depth, part 1
        if (i > 0 && char_O_deco_depth[i - 1] == char_O_deco_depth[i])
            k--;

        assert(i == 0 && k == 0 || i > 0 && k > 0);

        // mask to get rid of gas switch info
        di->stops[k].depth = char_O_deco_depth[i] & 0x7f;
        // ostc keeps deco time as unsigned char, aggregate by depth, part 2
        di->stops[k].time += char_O_deco_time[i];

        i++;
        k++;
    }
    di->n_stops = k;
    di->stops = realloc(di->stops, di->n_stops * sizeof(deco_stop_t));
    return di;
}

/**
 * Inject NDL information into deco information structure.
 *
 * The NDL information is gathered from internal OSTC firmware data
 * structtures.
 */
void _inject_ndl(deco_profile_t *profile, deco_info_t *deco_info) {
    unsigned short time;
    int i;

    assert(deco_info != NULL);

    if (char_O_nullzeit >= 240)
        deco_info->ndl = char_O_nullzeit;
    else if (char_O_nullzeit > 0
            || (char_O_nullzeit == 0 && deco_info->n_stops == 0)) {
        /* no NDL at end of the dive and no deco stops, still there is NDL
         * to calculate */
        time = 0;
        for (i = 0; i < profile->n_segments; i++)
            time += profile->segments[i].time;
        deco_info->ndl = time + char_O_nullzeit;
    }
}

/**
 * Print debug information about NDL, deco stops and tissues saturation.
 */
void _debug_deco(int time) {
    int i = 0;

    fprintf(stderr, "DEBUG: NDL (%ds): %dmin\n", time, char_O_nullzeit);

    if (char_O_deco_depth[0] != 0 || char_O_deco_time[0] != 0) {
        fprintf(stderr, "DEBUG: deco (%ds): ", time);
        while (i < NUM_STOPS &&
                (char_O_deco_depth[i] != 0 || char_O_deco_time[i] != 0)) {
            fprintf(stderr, "(%dm, %dmin), ",
                    char_O_deco_depth[i] & 0x7f,
                    char_O_deco_time[i]);
            i++;
        }
        fprintf(stderr, "\n");
    }

    fprintf(stderr, "DEBUG: n2 tissues (%ds): ", time);
    for (i = 0; i < NUM_COMP; i++)
        fprintf(stderr, "%f,", pres_tissue_N2[i]);
    fprintf(stderr, "\n");

    fprintf(stderr, "DEBUG: he tissues (%ds): ", time);
    for (i = 0; i < NUM_COMP; i++)
        fprintf(stderr, "%f,", pres_tissue_He[i]);
    fprintf(stderr, "\n");
}

/**
 * Reset OSTC internal structures to default values of OSTC firmware and
 * clear tissues.
 */
void _reset_deco_params() {
    int i;

    char_I_deco_model = 0;

    /* do not set char_I_step_is_1min as it gives deco calculations
     * different from the OSTC 2N itself; having it unset causes cpu to
     * burn bit longer, but gives more consistent deco stops, i.e. dive at
     * 45m for 25min with gas mixes air+ean50 vs. dive with gas mixes
     * air+ean50+o2 */
    char_I_step_is_1min = 0;

    char_I_N2_ratio = 79;
    char_I_He_ratio = 0;
    char_I_deco_distance = 0;

    for (i = 0; i < NUM_GAS; i++) {
        char_I_deco_gas_change[i] = 0;
        char_I_deco_N2_ratio[i] = 0;
        char_I_deco_He_ratio[i] = 0;
    }
    
    char_I_const_ppO2 = 0;
    
    char_I_GF_High_percentage = 90;
    char_I_GF_Low_percentage  = 30;
    
    char_I_saturation_multiplier = 110;
    char_I_desaturation_multiplier = 90;
    
    /* clear CNS and tissues */
    int_I_pres_respiration = DECO_STD_ATM;
    int_I_pres_surface = DECO_STD_ATM;
    deco_clear_CNS_fraction();
    deco_clear_tissue();

    /* firmware 2.0 - deco_clear_tissue sets the variable to 0 */
    char_I_depth_last_deco = 3;
}

/**
 * Set deco parameters before dive simulation.
 */
void _set_deco_params(deco_conf_t *conf) {
    int i;
    unsigned char o2, he;

    char_I_deco_model = 1; // ZH-L16C-GF
    char_I_depth_last_deco = conf->last_stop;

    char_I_GF_Low_percentage = conf->gf_low;
    char_I_GF_High_percentage = conf->gf_high;

    int_I_pres_surface = conf->pressure;
    char_I_current_gas = 1;

    for (i = 0; i < conf->n_gas_mixes; i++) {
        char_I_deco_gas_change[i] = conf->gas_list[i].depth;
        o2 = conf->gas_list[i].o2;
        he = conf->gas_list[i].he;
        char_I_deco_N2_ratio[i] = 100 - o2 - he;
        char_I_deco_He_ratio[i] = he;

        assert(char_I_deco_N2_ratio[i] + char_I_deco_He_ratio[i] + o2 == 100);
    }
    // set the bottom gas as this is not done when OSTC's
    // deco_calc_hauptroutine function is called
    char_I_N2_ratio = char_I_deco_N2_ratio[0];
    char_I_He_ratio = char_I_deco_He_ratio[0];

    // enable CCR mode
    if (conf->setpoints)
        char_I_const_ppO2 = (unsigned char) (conf->setpoints[0].ppO2 * 100.0 + 0.5);

#ifdef DECO_DEBUG
    fprintf(stderr, "DEBUG: CCR %d\n", char_I_const_ppO2);
#endif
}

deco_conf_t *deco_conf_new() {
    deco_conf_t *conf;
   
    conf = calloc(1, sizeof(deco_conf_t));
    if (conf == NULL)
        return NULL;

    conf->gf_low = 30;
    conf->gf_high = 85;
    conf->pressure = DECO_STD_ATM;
    deco_set_altitude(conf, 0);
    conf->last_stop = 3;

    return conf;
}

int deco_setpoint_add(deco_conf_t *conf, unsigned char depth, float ppO2) {
    int k = conf->n_setpoints;

    deco_errno = DECO_ERROR_OK;

    if (k == 1) {
        deco_errno = DECO_ERROR_SETPOINT_NUMBER;
        return -1;
    }

    if (conf->setpoints)
        conf->setpoints = realloc(conf->setpoints,
                (k + 1) * sizeof(deco_setpoint_t));
    else
        conf->setpoints = calloc(1, sizeof(deco_setpoint_t));

    conf->setpoints[k].depth = depth;
    conf->setpoints[k].ppO2 = ppO2;

    conf->n_setpoints += 1;

    return 0;
}

int deco_gas_add(deco_conf_t *conf, unsigned char depth, unsigned char o2,
        unsigned char he) {

    int k = conf->n_gas_mixes;

    deco_errno = DECO_ERROR_OK;

    if (k == NUM_GAS) {
        deco_errno = DECO_ERROR_GAS_NUMBER;
        return -1;
    }

    if (conf->gas_list)
        conf->gas_list = realloc(conf->gas_list,
                (k + 1) * sizeof(deco_gas_mix_t));
    else
        conf->gas_list = calloc(1, sizeof(deco_gas_mix_t));

    if (conf->gas_list == NULL) {
        deco_errno = DECO_ERROR_MEMORY;
        return -1;
    }

    conf->gas_list[k].o2 = o2;
    conf->gas_list[k].he = he;
    conf->gas_list[k].depth = depth;

    conf->n_gas_mixes += 1;

    return 0;
}

int deco_segment_add(deco_profile_t *profile, unsigned char depth,
        unsigned int time) {

    int k = profile->n_segments;

    deco_errno = DECO_ERROR_OK;

    if (depth > 130) {
        deco_errno = DECO_ERROR_MAX_DEPTH;
        return -1;
    }

    if (k == 1) {
        deco_errno = DECO_ERROR_NO_FEATURE;
        return -1;
    }

    if (profile->segments)
        profile->segments = realloc(profile->segments, (k + 1) * sizeof(deco_profile_segment_t));
    else
        profile->segments = calloc(1, sizeof(deco_profile_segment_t));

    if (profile->segments == NULL) {
        deco_errno = DECO_ERROR_MEMORY;
        return -1;
    }

    profile->segments[k].depth = depth;
    profile->segments[k].time = time;

    profile->n_segments += 1;

    return 0;
}

void deco_set_altitude(deco_conf_t *conf, unsigned short altitude) {
    float L = 0.0065;
    float T0 = 288.15;
    float g = 9.80665;
    float M = 0.0289644;
    float R = 8.31447;
    conf->pressure = DECO_STD_ATM
        * pow(1 - L * altitude / T0, (g * M) / (R * L));
}

deco_info_t *deco_calculate(deco_profile_t *profile, deco_conf_t *conf) {
    int i, k;
    deco_profile_segment_t segment;
    deco_info_t *deco_info;

    _reset_deco_params();
    _set_deco_params(conf);

    /* start dive */
    char_O_deco_status = 3;

    for (k = 0; k < profile->n_segments; k++) {
        segment = profile->segments[k];

        int_I_pres_respiration = segment.depth * DECO_METER_TO_MBAR
            + int_I_pres_surface;

        for (i=0; i < DECO_SIM_SPEED * segment.time; i++) {
            int_I_divemins = (i + DECO_SIM_SPEED) / DECO_SIM_SPEED;
            deco_calc_hauptroutine();
#ifdef DECO_DEBUG
            _debug_deco((i + 1) * 60 / DECO_SIM_SPEED);
#endif
        }
    }

#ifdef DECO_DEBUG
    fprintf(stderr, "DEBUG: ascent start\n");
#endif

    /* go through deco stops */
    while (char_O_deco_status) {
    	deco_calc_hauptroutine();
#ifdef DECO_DEBUG
        _debug_deco(-1);
#endif
    }

#ifdef DECO_DEBUG
    fprintf(stderr, "DEBUG: at surface\n");
#endif

     deco_info = _copy_deco_stops();
     _inject_ndl(profile, deco_info);
     return deco_info;
}

void deco_free(deco_profile_t * profile, deco_conf_t *conf, deco_info_t *info) {
    if (info != NULL) {
        free(info->stops);
        free(info);
    }

    if (conf != NULL) {
        free(conf->setpoints);
        free(conf->gas_list);
        free(conf);
    }

    if (profile != NULL) {
        free(profile->segments);
        free(profile);
    }
}

char *deco_error(const int errno) {
    return deco_errors[errno];
}
